@@ -1,7 +1,7 @@
 /*
  * event.c	Server event handling
  *
- * Version:	$Id: event.c,v 1.110 2008/06/06 13:10:27 aland Exp $
+ * Version:	$Id$
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -22,13 +22,12 @@
  */
 
 #include <freeradius-devel/ident.h>
-RCSID("$Id: event.c,v 1.110 2008/06/06 13:10:27 aland Exp $")
+RCSID("$Id$")
 
 #include <freeradius-devel/radiusd.h>
 #include <freeradius-devel/modules.h>
 #include <freeradius-devel/event.h>
 #include <freeradius-devel/detail.h>
-#include <freeradius-devel/radius_snmp.h>
 
 #include <freeradius-devel/rad_assert.h>
 
@@ -53,7 +52,7 @@ static fr_event_list_t	*el = NULL;
 static fr_packet_list_t	*pl = NULL;
 static int			request_num_counter = 0;
 static struct timeval		now;
-static time_t			start_time;
+time_t				fr_start_time;
 static int			have_children;
 static int			has_detail_listener = FALSE;
 static int			just_started = FALSE;
@@ -125,82 +124,6 @@ static void tv_add(struct timeval *tv, int usec_delay)
 	}
 }
 
-#ifdef WITH_SNMP
-static void snmp_inc_counters(REQUEST *request)
-{
-	if (!request->root->do_snmp) return;
-
-	if (request->master_state == REQUEST_COUNTED) return;
-
-	if ((request->listener->type != RAD_LISTEN_AUTH) &&
-	    (request->listener->type != RAD_LISTEN_ACCT)) return;
-
-	/*
-	 *	Update the SNMP statistics.
-	 *
-	 *	Note that we do NOT do this in a child thread.
-	 *	Instead, we update the stats when a request is
-	 *	deleted, because only the main server thread calls
-	 *	this function, which makes it thread-safe.
-	 */
-	switch (request->reply->code) {
-	case PW_AUTHENTICATION_ACK:
-		rad_snmp.auth.total_responses++;
-		rad_snmp.auth.total_access_accepts++;
-		if (request->client && request->client->auth) {
-			request->client->auth->accepts++;
-		}
-		break;
-
-	case PW_AUTHENTICATION_REJECT:
-		rad_snmp.auth.total_responses++;
-		rad_snmp.auth.total_access_rejects++;
-		if (request->client && request->client->auth) {
-			request->client->auth->rejects++;
-		}
-		break;
-
-	case PW_ACCESS_CHALLENGE:
-		rad_snmp.auth.total_responses++;
-		rad_snmp.auth.total_access_challenges++;
-		if (request->client && request->client->auth) {
-			request->client->auth->challenges++;
-		}
-		break;
-
-#ifdef WITH_ACCOUNTING
-	case PW_ACCOUNTING_RESPONSE:
-		rad_snmp.acct.total_responses++;
-		if (request->client && request->client->acct) {
-			request->client->acct->responses++;
-		}
-		break;
-#endif
-
-		/*
-		 *	No response, it must have been a bad
-		 *	authenticator.
-		 */
-	case 0:
-		if (request->packet->code == PW_AUTHENTICATION_REQUEST) {
-			rad_snmp.auth.total_bad_authenticators++;
-			if (request->client && request->client->auth) {
-				request->client->auth->bad_authenticators++;
-			}
-		}
-		break;
-
-	default:
-		break;
-	}
-
-	request->master_state = REQUEST_COUNTED;
-}
-#else
-#define snmp_inc_counters(_x)
-#endif
-
-
 static void remove_from_request_hash(REQUEST *request)
 {
 	if (!request->in_request_hash) return;
@@ -208,7 +131,7 @@ static void remove_from_request_hash(REQUEST *request)
 	fr_packet_list_yank(pl, request->packet);
 	request->in_request_hash = FALSE;
 
-	snmp_inc_counters(request);
+	request_stats_final(request);
 }
 
 
@@ -299,14 +222,6 @@ static int insert_into_proxy_hash(REQUEST *request)
 	PTHREAD_MUTEX_LOCK(&proxy_mutex);
 
 	request->home_server->currently_outstanding++;
-	request->home_server->total_requests_sent++;
-
-	/*
-	 *	On overflow, back up to ~0.
-	 */
-	if (!request->home_server->total_requests_sent) {
-		request->home_server->total_requests_sent--;
-	}
 
 	if (!fr_packet_list_id_alloc(proxy_list, request->proxy)) {
 		int found;
@@ -320,7 +235,7 @@ static int insert_into_proxy_hash(REQUEST *request)
 		proxy_listener = proxy_new_listener();
 		if (!proxy_listener) {
 			PTHREAD_MUTEX_UNLOCK(&proxy_mutex);
-			DEBUG2("ERROR: Failed to create a new socket for proxying requests.");
+			RDEBUG2("ERROR: Failed to create a new socket for proxying requests.");
 			return 0;
 		}
 
@@ -345,14 +260,14 @@ static int insert_into_proxy_hash(REQUEST *request)
 
 		if (!fr_packet_list_socket_add(proxy_list, proxy_listener->fd)) {
 			PTHREAD_MUTEX_UNLOCK(&proxy_mutex);
-			DEBUG2("ERROR: Failed to create a new socket for proxying requests.");
+			RDEBUG2("ERROR: Failed to create a new socket for proxying requests.");
 			return 0;
 
 		}
 
 		if (!fr_packet_list_id_alloc(proxy_list, request->proxy)) {
 			PTHREAD_MUTEX_UNLOCK(&proxy_mutex);
-			DEBUG2("ERROR: Failed to create a new socket for proxying requests.");
+			RDEBUG2("ERROR: Failed to create a new socket for proxying requests.");
 			return 0;
 		}
 
@@ -378,7 +293,7 @@ static int insert_into_proxy_hash(REQUEST *request)
 
 	if (proxy < 0) {
 		PTHREAD_MUTEX_UNLOCK(&proxy_mutex);
-		DEBUG2("ERROR: All sockets are full.");
+		RDEBUG2("ERROR: All sockets are full.");
 		return 0;
 	}
 
@@ -389,13 +304,13 @@ static int insert_into_proxy_hash(REQUEST *request)
 	if (!fr_packet_list_insert(proxy_list, &request->proxy)) {
 		fr_packet_list_id_free(proxy_list, request->proxy);
 		PTHREAD_MUTEX_UNLOCK(&proxy_mutex);
-		DEBUG2("ERROR: Failed to insert entry into proxy list");
+		RDEBUG2("ERROR: Failed to insert entry into proxy list");
 		return 0;
 	}
 
 	PTHREAD_MUTEX_UNLOCK(&proxy_mutex);
 
-	DEBUG3(" proxy: allocating destination %s port %d - Id %d",
+	RDEBUG3(" proxy: allocating destination %s port %d - Id %d",
 	       inet_ntop(request->proxy->dst_ipaddr.af,
 			 &request->proxy->dst_ipaddr.ipaddr, buf, sizeof(buf)),
 	       request->proxy->dst_port,
@@ -425,13 +340,13 @@ static void wait_for_proxy_id_to_expire(void *ctx)
 	if ((request->num_proxied_requests == request->num_proxied_responses) ||
 	    timercmp(&now, &request->when, >)) {
 		if (request->packet) {
-			DEBUG2("Cleaning up request %d ID %d with timestamp +%d",
+			RDEBUG2("Cleaning up request %d ID %d with timestamp +%d",
 			       request->number, request->packet->id,
-			       (unsigned int) (request->timestamp - start_time));
+			       (unsigned int) (request->timestamp - fr_start_time));
 		} else {
-			DEBUG2("Cleaning up request %d with timestamp +%d",
+			RDEBUG2("Cleaning up request %d with timestamp +%d",
 			       request->number,
-			       (unsigned int) (request->timestamp - start_time));
+			       (unsigned int) (request->timestamp - fr_start_time));
 		}
 		fr_event_delete(el, &request->ev);
 		remove_from_proxy_hash(request);
@@ -456,13 +371,13 @@ static void wait_for_child_to_die(void *ctx)
 		request->delay += (request->delay >> 1);
 		tv_add(&request->when, request->delay);
 
-		DEBUG2("Child is still stuck for request %d", request->number);
+		RDEBUG2("Child is still stuck for request %d", request->number);
 
 		INSERT_EVENT(wait_for_child_to_die, request);
 		return;
 	}
 
-	DEBUG2("Child is finally responsive for request %d", request->number);
+	RDEBUG2("Child is finally responsive for request %d", request->number);
 	remove_from_request_hash(request);
 
 #ifdef WITH_PROXY
@@ -493,9 +408,9 @@ static void cleanup_delay(void *ctx)
 	}
 #endif
 
-	DEBUG2("Cleaning up request %d ID %d with timestamp +%d",
+	RDEBUG2("Cleaning up request %d ID %d with timestamp +%d",
 	       request->number, request->packet->id,
-	       (unsigned int) (request->timestamp - start_time));
+	       (unsigned int) (request->timestamp - fr_start_time));
 
 	fr_event_delete(el, &request->ev);
 	request_free(&request);
@@ -509,7 +424,7 @@ static void reject_delay(void *ctx)
 	rad_assert(request->magic == REQUEST_MAGIC);
 	rad_assert(request->child_state == REQUEST_REJECT_DELAY);
 
-	DEBUG2("Sending delayed reject for request %d", request->number);
+	RDEBUG2("Sending delayed reject for request %d", request->number);
 
 	request->listener->send(request->listener, request);
 
@@ -532,6 +447,7 @@ static void revive_home_server(void *ctx)
 			 buffer, sizeof(buffer)),
 	       home->port);
 	home->currently_outstanding = 0;
+	home->revive_time = now;
 }
 
 
@@ -543,7 +459,7 @@ static void no_response_to_ping(void *ctx)
 
 	home->num_received_pings = 0;
 
-	DEBUG2("No response to status check %d from home server %s port %d",
+	RDEBUG2("No response to status check %d from home server %s port %d",
 	       request->number,
 	       inet_ntop(request->proxy->dst_ipaddr.af,
 			 &request->proxy->dst_ipaddr.ipaddr,
@@ -561,7 +477,7 @@ static void received_response_to_ping(REQUEST *request)
 
 	home->num_received_pings++;
 
-	DEBUG2("Received response to status check %d (%d in current sequence)",
+	RDEBUG2("Received response to status check %d (%d in current sequence)",
 	       request->number, home->num_received_pings);
 
 	if (home->num_received_pings < home->num_pings_to_alive) {
@@ -576,17 +492,18 @@ static void received_response_to_ping(REQUEST *request)
 	       request->proxy->dst_port);
 
 	if (!fr_event_delete(el, &home->ev)) {
-		DEBUG2("Hmm... no event for home server, WTF?");
+		RDEBUG2("Hmm... no event for home server, WTF?");
 	}
 
 	if (!fr_event_delete(el, &request->ev)) {
-		DEBUG2("Hmm... no event for request, WTF?");
+		RDEBUG2("Hmm... no event for request, WTF?");
 	}
 
 	wait_for_proxy_id_to_expire(request);
 
 	home->state = HOME_STATE_ALIVE;
 	home->currently_outstanding = 0;
+	home->revive_time = now;
 }
 
 
@@ -658,7 +575,7 @@ static void ping_home_server(void *ctx)
 	rad_assert(request->proxy_listener == NULL);
 
 	if (!insert_into_proxy_hash(request)) {
-		DEBUG2("ERROR: Failed inserting status check %d into proxy hash.  Discarding it.",
+		RDEBUG2("ERROR: Failed inserting status check %d into proxy hash.  Discarding it.",
 		       request->number);
 		request_free(&request);
 		return;
@@ -1034,7 +951,7 @@ static void wait_a_bit(void *ctx)
 #ifdef HAVE_PTHREAD_H
 		request->child_pid = NO_SUCH_CHILD_PID;
 #endif
-		snmp_inc_counters(request);
+		request_stats_final(request);
 		cleanup_delay(request);
 		return;
 
@@ -1043,7 +960,7 @@ static void wait_a_bit(void *ctx)
 #ifdef HAVE_PTHREAD_H
 		request->child_pid = NO_SUCH_CHILD_PID;
 #endif
-		snmp_inc_counters(request);
+		request_stats_final(request);
 
 	case REQUEST_PROXIED:
 		rad_assert(request->next_callback != NULL);
@@ -1070,7 +987,7 @@ static void wait_a_bit(void *ctx)
 	 *	mode, with no threads...
 	 */
 	if (!callback) {
-		DEBUG("WARNING: Internal sanity check failed in event handler for request %d: Discarding the request!", request->number);
+		RDEBUG("WARNING: Internal sanity check failed in event handler for request %d: Discarding the request!", request->number);
 		fr_event_delete(el, &request->ev);
 		remove_from_proxy_hash(request);
 		remove_from_request_hash(request);
@@ -1100,7 +1017,7 @@ static int process_proxy_reply(REQUEST *request)
 	 */
 	vp = pairfind(request->config_items, PW_POST_PROXY_TYPE);
 	if (vp) {
-		DEBUG2("  Found Post-Proxy-Type %s", vp->vp_strvalue);
+		RDEBUG2("  Found Post-Proxy-Type %s", vp->vp_strvalue);
 		post_proxy_type = vp->vp_integer;
 	}
 	
@@ -1110,9 +1027,9 @@ static int process_proxy_reply(REQUEST *request)
 		const char *old_server = request->server;
 		
 		request->server = request->home_pool->virtual_server;
-		DEBUG2(" server %s {", request->server);
+		RDEBUG2(" server %s {", request->server);
 		rcode = module_post_proxy(post_proxy_type, request);
-		DEBUG2(" }");
+		RDEBUG2(" }");
 		request->server = old_server;
 	} else {
 		rcode = module_post_proxy(post_proxy_type, request);
@@ -1200,7 +1117,7 @@ static int request_pre_handler(REQUEST *request)
 	}
 
 	if (rcode < 0) {
-		radlog(L_ERR, "%s Dropping packet without response.", librad_errstr);
+		radlog(L_ERR, "%s Dropping packet without response.", fr_strerror());
 		request->child_state = REQUEST_DONE;
 		return 0;
 	}
@@ -1232,12 +1149,12 @@ static int proxy_request(REQUEST *request)
 	char buffer[128];
 
 	if (request->home_server->server) {
-		DEBUG("ERROR: Cannot perform real proxying to a virtual server.");
+		RDEBUG("ERROR: Cannot perform real proxying to a virtual server.");
 		return 0;
 	}
 
 	if (!insert_into_proxy_hash(request)) {
-		DEBUG("ERROR: Failed inserting request into proxy hash.");
+		RDEBUG("ERROR: Failed inserting request into proxy hash.");
 		return 0;
 	}
 
@@ -1258,7 +1175,7 @@ static int proxy_request(REQUEST *request)
 	}
 	request->next_callback = no_response_to_proxied_request;
 
-	DEBUG2("Proxying request %d to home server %s port %d",
+	RDEBUG2("Proxying request %d to home server %s port %d",
 	       request->number,
 	       inet_ntop(request->proxy->dst_ipaddr.af,
 			 &request->proxy->dst_ipaddr.ipaddr,
@@ -1295,13 +1212,13 @@ static int proxy_to_virtual_server(REQUEST *request)
 	if (!request->home_server || !request->home_server->server) return 0;
 
 	if (request->parent) {
-		DEBUG2("WARNING: Cancelling proxy request to virtual server %s as this request was itself proxied.", request->home_server->server);
+		RDEBUG2("WARNING: Cancelling proxy request to virtual server %s as this request was itself proxied.", request->home_server->server);
 		return 0;
 	}
 
 	fake = request_alloc_fake(request);
 	if (!fake) {
-		DEBUG2("WARNING: Out of memory");
+		RDEBUG2("WARNING: Out of memory");
 		return 0;
 	}
 
@@ -1317,13 +1234,13 @@ static int proxy_to_virtual_server(REQUEST *request)
 #endif
 
 	} else {
-		DEBUG2("Unknown packet type %d", request->proxy->code);
+		RDEBUG2("Unknown packet type %d", request->proxy->code);
 		return 0;
 	}
 
-	DEBUG2(">>> Sending proxied request internally to virtual server.");
+	RDEBUG2(">>> Sending proxied request internally to virtual server.");
 	radius_handle_request(fake, fun);
-	DEBUG2("<<< Received proxied response from internal virtual server.");
+	RDEBUG2("<<< Received proxied response from internal virtual server.");
 
 	request->proxy_reply = fake->reply;
 	fake->reply = NULL;
@@ -1372,7 +1289,7 @@ static int successfully_proxied_request(REQUEST *request)
 
 	realm = realm_find2(realmname);
 	if (!realm) {
-		DEBUG2("ERROR: Cannot proxy to unknown realm %s", realmname);
+		RDEBUG2("ERROR: Cannot proxy to unknown realm %s", realmname);
 		return 0;
 	}
 
@@ -1392,14 +1309,14 @@ static int successfully_proxied_request(REQUEST *request)
 	}
 
 	if (!pool) {
-		DEBUG2(" WARNING: Cancelling proxy to Realm %s, as the realm is local.",
+		RDEBUG2(" WARNING: Cancelling proxy to Realm %s, as the realm is local.",
 		       realmname);
 		return 0;
 	}
 
 	home = home_server_ldb(realmname, pool, request);
 	if (!home) {
-		DEBUG2("ERROR: Failed to find live home server for realm %s",
+		RDEBUG2("ERROR: Failed to find live home server for realm %s",
 		       realmname);
 		return -1;
 	}
@@ -1421,7 +1338,7 @@ static int successfully_proxied_request(REQUEST *request)
 	    (home->ipaddr.af == AF_INET) &&
 	    (request->packet->src_ipaddr.af == AF_INET) &&
 	    (home->ipaddr.ipaddr.ip4addr.s_addr == request->packet->src_ipaddr.ipaddr.ip4addr.s_addr)) {
-		DEBUG2("    rlm_realm: Packet came from realm %s, proxy cancelled", realmname);
+		RDEBUG2("    rlm_realm: Packet came from realm %s, proxy cancelled", realmname);
 		return 0;
 	}
 #endif
@@ -1537,7 +1454,7 @@ static int successfully_proxied_request(REQUEST *request)
 	 */
 	vp = pairfind(request->config_items, PW_PRE_PROXY_TYPE);
 	if (vp) {
-		DEBUG2("  Found Pre-Proxy-Type %s", vp->vp_strvalue);
+		RDEBUG2("  Found Pre-Proxy-Type %s", vp->vp_strvalue);
 		pre_proxy_type = vp->vp_integer;
 	}
 
@@ -1547,9 +1464,9 @@ static int successfully_proxied_request(REQUEST *request)
 		const char *old_server = request->server;
 		
 		request->server = request->home_pool->virtual_server;
-		DEBUG2(" server %s {", request->server);
+		RDEBUG2(" server %s {", request->server);
 		rcode = module_pre_proxy(pre_proxy_type, request);
-		DEBUG2(" }");
+		RDEBUG2(" }");
 			request->server = old_server;
 	} else {
 		rcode = module_pre_proxy(pre_proxy_type, request);
@@ -1591,7 +1508,7 @@ static int successfully_proxied_request(REQUEST *request)
 	}
 
 	if (!proxy_request(request)) {
-		DEBUG("ERROR: Failed to proxy request %d", request->number);
+		RDEBUG("ERROR: Failed to proxy request %d", request->number);
 		return -1;
 	}
 	
@@ -1609,7 +1526,7 @@ static void request_post_handler(REQUEST *request)
 	if ((request->master_state == REQUEST_STOP_PROCESSING) ||
 	    (request->parent &&
 	     (request->parent->master_state == REQUEST_STOP_PROCESSING))) {
-		DEBUG2("Request %d was cancelled.", request->number);
+		RDEBUG2("Request %d was cancelled.", request->number);
 #ifdef HAVE_PTHREAD_H
 		request->child_pid = NO_SUCH_CHILD_PID;
 #endif
@@ -1665,7 +1582,7 @@ static void request_post_handler(REQUEST *request)
 	 *	here!
 	 */
 	if (request->packet->dst_port == 0) {
-		/* FIXME: DEBUG going to the next request */
+		/* FIXME: RDEBUG going to the next request */
 #ifdef HAVE_PTHREAD_H
 		request->child_pid = NO_SUCH_CHILD_PID;
 #endif
@@ -1695,11 +1612,11 @@ static void request_post_handler(REQUEST *request)
 			vp = pairfind(request->config_items,
 				      PW_RESPONSE_PACKET_TYPE);
 			if (!vp) {
-				DEBUG2("There was no response configured: rejecting request %d",
+				RDEBUG2("There was no response configured: rejecting request %d",
 				       request->number);
 				request->reply->code = PW_AUTHENTICATION_REJECT;
 			} else if (vp->vp_integer == 256) {
-				DEBUG2("Not responding to request %d",
+				RDEBUG2("Not responding to request %d",
 				       request->number);
 
 			} else {
@@ -1730,7 +1647,7 @@ static void request_post_handler(REQUEST *request)
 			when.tv_sec += request->root->reject_delay;
 
 			if (timercmp(&when, &request->next_when, >)) {
-				DEBUG2("Delaying reject of request %d for %d seconds",
+				RDEBUG2("Delaying reject of request %d for %d seconds",
 				       request->number,
 				       request->root->reject_delay);
 				request->next_when = when;
@@ -1806,6 +1723,7 @@ static void request_post_handler(REQUEST *request)
 			pairfree(&request->proxy_reply->vps);
 		}
 
+#if 0
 		/*
 		 *	We're not tracking responses from the home
 		 *	server, we can therefore free this memory in
@@ -1816,10 +1734,11 @@ static void request_post_handler(REQUEST *request)
 			rad_free(&request->proxy_reply);
 			request->home_server = NULL;
 		}
+#endif
 	}
 #endif
 
-	DEBUG2("Finished request %d.", request->number);
+	RDEBUG2("Finished request %d.", request->number);
 
 	request->child_state = child_state;
 
@@ -1836,8 +1755,8 @@ static void received_retransmit(REQUEST *request, const RADCLIENT *client)
 	char buffer[128];
 #endif
 
-	RAD_SNMP_TYPE_INC(request->listener, total_dup_requests);
-	RAD_SNMP_CLIENT_INC(request->listener, client, dup_requests);
+	RAD_STATS_TYPE_INC(request->listener, total_dup_requests);
+	RAD_STATS_CLIENT_INC(request->listener, client, total_dup_requests);
 
 	switch (request->child_state) {
 	case REQUEST_QUEUED:
@@ -1885,7 +1804,7 @@ static void received_retransmit(REQUEST *request, const RADCLIENT *client)
 
 			home = home_server_ldb(NULL, request->home_pool, request);
 			if (!home) {
-				DEBUG2("Failed to find live home server for request %d", request->number);
+				RDEBUG2("Failed to find live home server for request %d", request->number);
 			no_home_servers:
 				/*
 				 *	Do post-request processing,
@@ -1922,7 +1841,7 @@ static void received_retransmit(REQUEST *request, const RADCLIENT *client)
 			 *	Try to proxy the request.
 			 */
 			if (!proxy_request(request)) {
-				DEBUG("ERROR: Failed to re-proxy request %d", request->number);
+				RDEBUG("ERROR: Failed to re-proxy request %d", request->number);
 				goto no_home_servers;
 			}
 
@@ -1936,7 +1855,7 @@ static void received_retransmit(REQUEST *request, const RADCLIENT *client)
 			return;
 		} /* else the home server is still alive */
 
-		DEBUG2("Sending duplicate proxied request to home server %s port %d - ID: %d",
+		RDEBUG2("Sending duplicate proxied request to home server %s port %d - ID: %d",
 		       inet_ntop(request->proxy->dst_ipaddr.af,
 				 &request->proxy->dst_ipaddr.ipaddr,
 				 buffer, sizeof(buffer)),
@@ -1949,7 +1868,7 @@ static void received_retransmit(REQUEST *request, const RADCLIENT *client)
 #endif
 
 	case REQUEST_REJECT_DELAY:
-		DEBUG2("Waiting to send Access-Reject "
+		RDEBUG2("Waiting to send Access-Reject "
 		       "to client %s port %d - ID: %d",
 		       client->shortname,
 		       request->packet->src_port, request->packet->id);
@@ -1957,7 +1876,7 @@ static void received_retransmit(REQUEST *request, const RADCLIENT *client)
 
 	case REQUEST_CLEANUP_DELAY:
 	case REQUEST_DONE:
-		DEBUG2("Sending duplicate reply "
+		RDEBUG2("Sending duplicate reply "
 		       "to client %s port %d - ID: %d",
 		       client->shortname,
 		       request->packet->src_port, request->packet->id);
@@ -2149,6 +2068,11 @@ int received_request(rad_listen_t *listener,
 	request->number = request_num_counter++;
 	request->priority = listener->type;
 
+	/*
+	 *	Status-Server packets go to the head of the queue.
+	 */
+	if (request->packet->code == PW_STATUS_SERVER) request->priority = 0;
+
 	/*
 	 *	Set virtual server identity
 	 */
@@ -2251,10 +2175,11 @@ REQUEST *received_proxy_response(RADIUS_PACKET *packet)
 	home = request->home_server;
 
 	gettimeofday(&now, NULL);
+
 	home->state = HOME_STATE_ALIVE;
 
 	if (request->reply && request->reply->code != 0) {
-		DEBUG2("We already replied to this request.  Discarding response from home server.");
+		RDEBUG2("We already replied to this request.  Discarding response from home server.");
 		rad_free(&packet);
 		return NULL;
 	}
@@ -2267,7 +2192,7 @@ REQUEST *received_proxy_response(RADIUS_PACKET *packet)
 		if (memcmp(request->proxy_reply->vector,
 			   packet->vector,
 			   sizeof(request->proxy_reply->vector)) == 0) {
-			DEBUG2("Discarding duplicate reply from home server %s port %d  - ID: %d for request %d",
+			RDEBUG2("Discarding duplicate reply from home server %s port %d  - ID: %d for request %d",
 			       inet_ntop(packet->src_ipaddr.af,
 					 &packet->src_ipaddr.ipaddr,
 					 buffer, sizeof(buffer)),
@@ -2279,13 +2204,27 @@ REQUEST *received_proxy_response(RADIUS_PACKET *packet)
 			 *	reply, which doesn't match the old
 			 *	one.  Delete it.
 			 */
-			DEBUG2("Ignoring conflicting proxy reply");
+			RDEBUG2("Ignoring conflicting proxy reply");
 		}
 
 		/* assert that there's an event queued for request? */
 		rad_free(&packet);
 		return NULL;
 	}
+#ifdef WITH_STATS
+	/*
+	 *	The average includes our time to receive packets and
+	 *	look them up in the hashes, which should be the same
+	 *	for all packets.
+	 *
+	 *	We update the response time only for the FIRST packet
+	 *	we receive.
+	 */
+	else if (home->ema.window > 0) {
+		radius_stats_ema(&home->ema, &now, &request->proxy_when);
+	}
+#endif
+
 
 	switch (request->child_state) {
 	case REQUEST_QUEUED:
@@ -2375,6 +2314,42 @@ REQUEST *received_proxy_response(RADIUS_PACKET *packet)
 }
 #endif
 
+void event_new_fd(rad_listen_t *this)
+{
+	char buffer[1024];
+	
+	this->print(this, buffer, sizeof(buffer));
+	
+	if (this->status == RAD_LISTEN_STATUS_INIT) {
+		DEBUG2(" ... adding new socket %s", buffer);
+		
+		if (!fr_event_fd_insert(el, 0, this->fd,
+					event_socket_handler, this)) {
+			radlog(L_ERR, "Failed remembering handle for proxy socket!");
+			exit(1);
+		}
+		
+		this->status = RAD_LISTEN_STATUS_KNOWN;
+		return;
+	}
+	
+	if (this->status == RAD_LISTEN_STATUS_CLOSED) {
+		DEBUG2(" ... closing socket %s", buffer);
+		
+		fr_event_fd_delete(el, 0, this->fd);
+		this->status = RAD_LISTEN_STATUS_FINISH;
+		
+		/*
+		 *	Close the fd AFTER fixing up the requests and
+		 *	listeners, so that they don't send/recv on the
+		 *	wrong socket (if someone manages to open
+		 *	another one).
+		 */
+		close(this->fd);
+		this->fd = -1;
+	}
+}
+
 #ifdef WITH_DETAIL
 static void event_detail_timer(void *ctx)
 {
@@ -2441,7 +2416,7 @@ static void handle_signal_self(int flag)
 			tv_add(&when, delay);
 
 			if (delay > 100000) {
-				DEBUG2("Delaying next detail event for %d.%01u seconds.",
+				DEBUG("Delaying next detail event for %d.%01u seconds.",
 				       delay / USEC, (delay % USEC) / 100000);
 			}
 
@@ -2460,13 +2435,9 @@ static void handle_signal_self(int flag)
 		for (this = mainconfig.listen;
 		     this != NULL;
 		     this = this->next) {
-			if (this->type != RAD_LISTEN_PROXY) continue;
-			
-			if (!fr_event_fd_insert(el, 0, this->fd,
-						event_socket_handler, this)) {
-				radlog(L_ERR, "Failed remembering handle for proxy socket!");
-				exit(1);
-			}
+			if (this->status == RAD_LISTEN_STATUS_KNOWN) continue;
+
+			event_new_fd(this);
 		}
 	}
 }
@@ -2551,7 +2522,7 @@ static void event_socket_handler(fr_event_list_t *xel, UNUSED int fd,
  *	This function is called periodically to see if any FD's are
  *	available for reading.
  */
-static void event_poll_fds(UNUSED void *ctx)
+static void event_poll_detail(UNUSED void *ctx)
 {
 	int rcode;
 	RAD_REQUEST_FUNP fun;
@@ -2564,6 +2535,8 @@ static void event_poll_fds(UNUSED void *ctx)
 	when.tv_sec += 1;
 
 	for (this = mainconfig.listen; this != NULL; this = this->next) {
+		if (this->type != RAD_LISTEN_DETAIL) continue;
+
 		if (this->fd >= 0) continue;
 
 		/*
@@ -2580,40 +2553,12 @@ static void event_poll_fds(UNUSED void *ctx)
 		if (!thread_pool_addrequest(request, fun)) {
 			request->child_state = REQUEST_DONE;
 		}
-
-		/*
-		 *	We have an FD.  Start watching it.
-		 */
-		if (this->fd >= 0) {
-#ifdef WITH_DETAIL
-			/*
-			 *	... unless it's a detail file.  In
-			 *	that case, we rely on the signal to
-			 *	self to know when to continue
-			 *	processing the detail file.
-			 */
-			if (this->type == RAD_LISTEN_DETAIL) continue;
-#endif
-
-			/*
-			 *	FIXME: this should be SNMP handler,
-			 *	and we should do SOMETHING when the
-			 *	fd is closed!
-			 */
-			if (!fr_event_fd_insert(el, 0, this->fd,
-						event_socket_handler, this)) {
-				char buffer[256];
-				
-				this->print(this, buffer, sizeof(buffer));
-				rad_panic("Failed creating handler for snmp");
-			}
-		}
 	}
 
 	/*
 	 *	Reset the poll.
 	 */
-	if (!fr_event_insert(el, event_poll_fds, NULL,
+	if (!fr_event_insert(el, event_poll_detail, NULL,
 			     &when, NULL)) {
 		radlog(L_ERR, "Failed creating handler");
 		exit(1);
@@ -2670,12 +2615,11 @@ static void event_status(struct timeval *wake)
  */
 int radius_event_init(CONF_SECTION *cs, int spawn_flag)
 {
-	int has_snmp_listener = FALSE;
 	rad_listen_t *this, *head = NULL;
 
 	if (el) return 0;
 
-	time(&start_time);
+	time(&fr_start_time);
 
 	el = fr_event_list_create(event_status);
 	if (!el) return 0;
@@ -2724,7 +2668,7 @@ int radius_event_init(CONF_SECTION *cs, int spawn_flag)
 #endif
 
 	if (check_config) {
-		DEBUG2("%s: #### Skipping IP addresses and Ports ####",
+		DEBUG("%s: #### Skipping IP addresses and Ports ####",
 		       mainconfig.name);
 		return 1;
 	}
@@ -2768,7 +2712,7 @@ int radius_event_init(CONF_SECTION *cs, int spawn_flag)
 	}
 #endif
 
-	DEBUG2("%s: #### Opening IP addresses and Ports ####",
+       DEBUG("%s: #### Opening IP addresses and Ports ####",
 	       mainconfig.name);
 
 	if (listen_init(cs, &head) < 0) {
@@ -2793,13 +2737,6 @@ int radius_event_init(CONF_SECTION *cs, int spawn_flag)
 			break;
 #endif
 
-#ifdef WITH_SNMP
-		case RAD_LISTEN_SNMP:
-			DEBUG("Listening on SNMP %s", buffer);
-			has_snmp_listener = TRUE;
-			break;
-#endif
-
 #ifdef WITH_PROXY
 		case RAD_LISTEN_PROXY:
 			rad_assert(proxy_fds[this->fd & 0x1f] == -1);
@@ -2821,8 +2758,8 @@ int radius_event_init(CONF_SECTION *cs, int spawn_flag)
 
 		/*
 		 *	The file descriptor isn't ready.  Poll for
-		 *	when it will become ready.  This is for SNMP
-		 *	and detail file fd's.
+		 *	when it will become ready.  This is for the
+		 *	detail file fd's.
 		 */
 		if (this->fd < 0) {
 			continue;
@@ -2842,15 +2779,17 @@ int radius_event_init(CONF_SECTION *cs, int spawn_flag)
 			       buffer);
 			exit(1);
 		}
+
+		this->status = RAD_LISTEN_STATUS_KNOWN;
 	}
 
-	if (has_detail_listener || has_snmp_listener) {
+	if (has_detail_listener) {
 		struct timeval when;
 		
 		gettimeofday(&when, NULL);
 		when.tv_sec += 1;
 		
-		if (!fr_event_insert(el, event_poll_fds, NULL,
+		if (!fr_event_insert(el, event_poll_detail, NULL,
 				     &when, NULL)) {
 			radlog(L_ERR, "Failed creating handler");
 			exit(1);
@@ -2937,15 +2876,17 @@ int radius_event_process(void)
 
 void radius_handle_request(REQUEST *request, RAD_REQUEST_FUNP fun)
 {
+	request->options = RAD_REQUEST_OPTION_DEBUG2;
+
 	if (request_pre_handler(request)) {
 		rad_assert(fun != NULL);
 		rad_assert(request != NULL);
 		
-		if (request->server) DEBUG("server %s {",
+		if (request->server) RDEBUG("server %s {",
 					     request->server); 
 		fun(request);
 
-		if (request->server) DEBUG("} # server %s",
+		if (request->server) RDEBUG("} # server %s",
 					     request->server);
 
 		request_post_handler(request);
